////////////////////////////////////////////////////////////////////////////////
//      _    _  ____            ____    // Core.pas: unit Core                //
//     / \  | |/ ___| ___ _ __ |___ \   //------------------------------------//
//    / _ \ | | |  _ / _ \ '_ \  __) |  // Routinen zur Berechnung eines      //
//   / ___ \| | |_| |  __/ | | |/ __/   // Bildes und zur Konstruktion der    //
//  /_/   \_\_|\____|\___|_| |_|_____|  // Zieldateinamen                     //
//                                      //                                    //
////////////////////////////////////////////////////////////////////////////////
unit Core;
interface
uses Graphics, Datacore;


type TImageType=(itPreview,itThumbnail);  // Bildtypen

function PreprocessImage(const ImageRec:TImageRecord):TBitmap;
                  // ldt und vorverarbeitet (Median, Equalize, Rotate) ein Bild
function ScaleImage(const ImageRec:TImageRecord; Raw:TBitmap;
                    ImageType:TImageType):TBitmap;
                 // skaliert ein Bild und fgt (bei Thumbnails) den Rahmen hinzu
function TestCompressImage(Uncompressed:TBitmap; ImageType:TImageType;
                           out Size:integer):TBitmap;
                // komprimiert ein Bild im Speicher, dekomprimiert es wieder und
                // gibt die Gre aus (zu Vorschauzwecken)
function CompressImage(Uncompressed:TBitmap; const FileName:string;
                       ImageType:TImageType):boolean;
                                           // komprimiert ein Bild in eine Datei


function BuildNumber(Number:integer):string;
   // fllt die Nummer mit fhrenden Nullen auf, wie im Hauptfenster eingestellt
function BuildFileName(Number:integer; Template:string):string;
           // konstruiert einen Dateivornamen aus einem Template im Hauptfenster


implementation
uses SysUtils, Classes, ScaleBMP, RotateBMP, Histogram, JPEG, Main;


////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////// Rendering-Helferlein
////////////////////////////////////////////////////////////////////////////////

///// MedianFilter: Filtert ein Bitmap mit einem Median-Cut-Filter
procedure MedianFilter(BMP:TBitmap);
var X,Y,MaxX:integer;
    L:PByteArray; CP,NP:^byte;
    LR,LG,LB,CR,CG,CB,NR,NG,NB:integer;
  ///// Med: liefert den Median aus drei Integerzahlen
  function Med(a,b,c:integer):integer;
  begin
    if a<b then begin
      if b<c then Result:=b else
      if a<c then Result:=c else
                  Result:=a;
    end else begin
      if b>c then Result:=b else
      if a<c then Result:=a else
                  Result:=c;
    end;
  end;
begin
  BMP.PixelFormat:=pf24bit;
  MaxX:=BMP.Width-3;
  // zeilenweises Vorgehen mit herrlich verworrender ScanLine-Technik
  for Y:=(BMP.Height-1) downto 0 do begin
    L:=BMP.ScanLine[Y];
    LB:=L[0]; LG:=L[1]; LR:=L[2];
    CB:=L[3]; CG:=L[4]; CR:=L[5];
    CP:=@L[3]; NP:=@L[6];
    for X:=MaxX downto 0 do begin
      // drei Farbebenen unabhngig filtern
      NB:=NP^; CP^:=Med(LB,CB,NB); LB:=CB; CB:=NB; inc(CP); inc(NP);
      NG:=NP^; CP^:=Med(LG,CG,NG); LG:=CG; CG:=NG; inc(CP); inc(NP);
      NR:=NP^; CP^:=Med(LR,CR,NR); LR:=CR; CR:=NR; inc(CP); inc(NP);
    end;
  end;
end;


///// RenderFrame: fgt einem Bitmap einen 3D-Rahmen hinzu
procedure RenderFrame(BMP:TBitmap; FrameSize:integer);
var MaxX,MaxY:integer;
    T,X,Y:integer;
    FrameMul,FrameDiv:integer;
    vR,vG,vB:integer;
    L:PByteArray; LIndex:integer;
  // GetPixel: liest ein Pixel aus dem Bitmap (mit ScanLine-Cache)
  procedure GetPixel(X,Y:integer);
  var O:integer;
  begin
    if LIndex<>Y then begin L:=BMP.ScanLine[Y]; LIndex:=Y; end;
    O:=3*X;
    vB:=L[O]; vG:=L[O+1]; vR:=L[O+2];
  end;
  // PutPixel: schreibt ein Pixel in das Bitmap (mit ScanLine-Cache)
  procedure PutPixel(X,Y:integer);
  var O:integer;
  begin
    if LIndex<>Y then begin L:=BMP.ScanLine[Y]; LIndex:=Y; end;
    O:=3*X;
    L[O]:=vB; L[O+1]:=vG; L[O+2]:=vR;
  end;
  // Lighten: holt sich ein Pixel, hellt es auf und schreibt es zurck
  procedure Lighten(X,Y:integer);
  begin
    GetPixel(X,Y);
    vR:=((vR XOR 255)*FrameMul DIV FrameDiv) XOR 255;
    vG:=((vG XOR 255)*FrameMul DIV FrameDiv) XOR 255;
    vB:=((vB XOR 255)*FrameMul DIV FrameDiv) XOR 255;
    PutPixel(X,Y);
  end;
  // Darken: holt sich ein Pixel, dunkelt es ab und schreibt es zurck
  procedure Darken(X,Y:integer);
  begin
    GetPixel(X,Y);
    vR:=vR*FrameMul DIV FrameDiv;
    vG:=vG*FrameMul DIV FrameDiv;
    vB:=vB*FrameMul DIV FrameDiv;
    PutPixel(X,Y);
  end;
begin
  // Vorberechnungen
  if FrameSize<1 then exit;
  MaxX:=BMP.Width-1;
  MaxY:=BMP.Height-1;
  if (FrameSize>(MaxX SHR 1)) OR (FrameSize>(MaxY SHR 1)) then exit;
                                                     // zu groe Frames ablehnen
  FrameDiv:=FrameSize+1;
  LIndex:=-1;
  // Frame von auen nach innen rendern
  for T:=0 to (FrameSize-1) do begin
    FrameMul:=T+1;
    // an allen vier Seiten abschrgen
    for X:=T to MaxX-T do Lighten(X,T);
    for Y:=T+1 to MaxY-T do Lighten(T,Y);
    for X:=T+1 to MaxX-T do Darken(X,MaxY-T);
    for Y:=T+1 to MaxY-T-1 do Darken(MaxX-T,Y);
  end;
end;


////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////// Bildverarbeitung
////////////////////////////////////////////////////////////////////////////////

///// PreprocessImage: Bild laden und vorverarbeiten
function PreprocessImage(const ImageRec:TImageRecord):TBitmap;
var JPEG:TJPEGImage; Raw:TBitmap;
    EqualizeMode:TEqualizeMode;
begin
  JPEG:=nil; Raw:=nil; Result:=nil;
  try
    if LowerCase(ExtractFileExt(ImageRec.PathName))='.bmp' then begin
      // Phase 1, Variante 1: Bitmap lesen
      Raw:=TBitmap.Create;
      Raw.LoadFromFile(ImageRec.PathName);
    end else begin
      // Phase 1, Variante 2: JPEG lesen und Raw-Image erzeugen
      JPEG:=TJPEGImage.Create;
      JPEG.Performance:=jpBestQuality;
      JPEG.LoadFromFile(ImageRec.PathName);
      Raw:=TBitmap.Create;
      Raw.Assign(JPEG);
      FreeAndNil(JPEG);
    end;
    // Phase 2: Preprocessing
    if MainForm.CMedianFilter.Checked then MedianFilter(Raw);
    if MainForm.CEnableEqualize.Checked then begin
      if MainForm.CEqualizeRGB.Checked
        then EqualizeMode:=emRGB
        else EqualizeMode:=emGray;
      Equalize(Raw,EqualizeMode,MainForm.CEqualizeStrength.Position*0.001);
    end;
    // Phase 3: Image drehen
    case ImageRec.Rotation of
      rot90CW:Result:=Rotate90CW(Raw);
      rot90CCW:Result:=Rotate90CCW(Raw);
      rot180:Result:=Rotate180(Raw);
    else
      Result:=Rotate0(Raw);
    end;
    Raw.FreeImage;
    FreeAndNil(Raw);
  except
    // im Fehlerfall alles freigeben und nil zurckliefern
    if Assigned(JPEG) then JPEG.Free;
    if Assigned(Raw) then Raw.Free;
    if Assigned(Result) then FreeAndNil(Result);
  end;
end;


///// ScaleImage: bringt ein (bereits vorverarbeitetes) Bild auf Vorschau- oder
///// Thumbnailgre und fgt im letzteren Fall Rahmen hinzu
function ScaleImage(const ImageRec:TImageRecord; Raw:TBitmap;
                    ImageType:TImageType):TBitmap;
var OldX,OldY,NewX,NewY,SizeX,SizeY:integer;
begin
  Result:=nil;
  if not Assigned(Raw) then exit;
  try
    // Phase 4a: Gre berechnen
    if ImageType=itPreview then begin
      SizeX:=MainForm.CPSizeX.Position;
      SizeY:=MainForm.CPSizeY.Position;
    end else begin
      SizeX:=MainForm.CTSizeX.Position;
      SizeY:=MainForm.CTSizeY.Position;
    end;
    OldX:=Raw.Width;
    OldY:=Raw.Height;
    if OldX>SizeX then begin
      NewX:=SizeX;
      NewY:=SizeX*OldY DIV OldX;
    end else begin
      NewX:=OldX;
      NewY:=OldY;
    end;
    if NewY>SizeY then begin
      NewY:=SizeY;
      NewX:=NewY*OldX DIV OldY;
    end;
    // Phase 4b: Image auf Zielgre skalieren
    Result:=TBitmap.Create;
    Result.PixelFormat:=pf24bit;
    Result.Width:=NewX;
    Result.Height:=NewY;
    Scale(Raw,Result);
    // Phase 5: Frame hinzufgen
    if (ImageType=itThumbnail) AND MainForm.CEnableFrame.Checked then
      RenderFrame(Result,MainForm.CFrameSize.Position);
  except
    // im Fehlerfall nil zurckliefern
    if Assigned(Result) then FreeAndNil(Result);
  end;
end;


////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////// Bildkompression
////////////////////////////////////////////////////////////////////////////////

///// TestCompressImage: komprimiert ein Bild probehalber mit den angegebenen
///// Einstellungen und liefert die komprimierte Gre sowie das wieder ent-
///// komprimierte Bild zurck
function TestCompressImage(Uncompressed:TBitmap; ImageType:TImageType;
                           out Size:integer):TBitmap;
var Stream:TMemoryStream;
    JPEG:TJPEGImage;
begin
  Result:=nil; Stream:=nil; JPEG:=nil;
  if not Assigned(Uncompressed) then exit;
  try
    Stream:=TMemoryStream.Create;
    JPEG:=TJPEGImage.Create;
    JPEG.Assign(Uncompressed);
    // JPEG-Kompressor konfigurieren
    JPEG.Grayscale:=false;
    if ImageType=itPreview then begin
      JPEG.CompressionQuality:=MainForm.CPQuality.Position;
      JPEG.ProgressiveEncoding:=MainForm.CPProgressive.Checked;
    end else begin
      JPEG.CompressionQuality:=MainForm.CTQuality.Position;
      JPEG.ProgressiveEncoding:=MainForm.CTProgressive.Checked;
    end;
    // JPEG im MemoryStream abspeichern und dort Gre ermitteln
    JPEG.SaveToStream(Stream);
    Size:=Stream.Position;
    // Stream zurckspulen, JPEG-Dekompressor konfigurieren
    Stream.Seek(0,soFromBeginning);
    JPEG.Performance:=jpBestQuality;
    // Bild auspacken und in Bitmap kopieren
    JPEG.LoadFromStream(Stream);
    Stream.Free;
    JPEG.DIBNeeded;
    Result:=TBitmap.Create;
    Result.Assign(JPEG);
    JPEG.Free;
  except
    // im Fehlerfall alles freigeben und nil zurckliefern
    if Assigned(Stream) then Stream.Free;
    if Assigned(JPEG) then JPEG.Free;
    if Assigned(Result) then FreeAndNil(Result);
  end;
end;


///// CompressImage: komprimiert ein Bild probehalber mit den angegebenen
///// Einstellungen in eine Datei
function CompressImage(Uncompressed:TBitmap; const FileName:string;
                       ImageType:TImageType):boolean;
var JPEG:TJPEGImage;
begin
  Result:=false; JPEG:=nil;
  if not Assigned(Uncompressed) then exit;
  try
    JPEG:=TJPEGImage.Create;
    JPEG.Assign(Uncompressed);
    // JPEG-Kompressor konfigurieren
    JPEG.Grayscale:=false;
    if ImageType=itPreview then begin
      JPEG.CompressionQuality:=MainForm.CPQuality.Position;
      JPEG.ProgressiveEncoding:=MainForm.CPProgressive.Checked;
    end else begin
      JPEG.CompressionQuality:=MainForm.CTQuality.Position;
      JPEG.ProgressiveEncoding:=MainForm.CTProgressive.Checked;
    end;
    // JPEG in Datei speichern
    JPEG.SaveToFile(FileName);
    JPEG.Free;
  except
    // im Fehlerfall JPEG-Objekt freigeben
    if Assigned(JPEG) then JPEG.Free;
  end;
  Result:=true;
end;


////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////// Dateinamengenerator
////////////////////////////////////////////////////////////////////////////////

///// BuildNumber: wandelt eine Zahl in einen String um, ggf. mit fhrenden 0en
function BuildNumber(Number:integer):string;
var Digits,i:integer;
begin
  Digits:=MainForm.CDigits.Position;
  Str(Number:Digits,Result);
  for i:=1 to Digits do
    if Result[i]=#32 then Result[i]:=#48;
end;

///// BuildFileName: baut die Nummer passend in ein Dateinamen-Template ein
function BuildFileName(Number:integer; Template:string):string;
var i:integer;
begin
  // Sterne und Fragezeichen zu Prozentzeichen wandeln
  Template:=StringReplace(StringReplace(Template,
            '?','%',[rfReplaceAll]),'*','%',[rfReplaceAll]);
  // Prozentzeichen suchen
  i:=pos('%',Template);
  // keins da -> am Ende des Strings eins "dazudenken"
  if i<1 then i:=length(Template)+1;
  // Zieldateiname konstruieren
  Result:=copy(Template,1,i-1)+BuildNumber(Number)+copy(Template,i+1,1024);
end;

end.
