unit Histogram;
// Histogram Stretch Toolkit (C) 2003 Martin J. Fiedler <martin.fiedler@gmx.net>
// This is free software. You may freely redistribute and modify it.
interface
uses Graphics;

type TEqualizeMode=(emGray,emRGB);
procedure Equalize(BMP:TBitmap; Mode:TEqualizeMode; Quantile:real);
// Simple operation: Pass a TBitmap to Equalize(). All dark pixels that occopy
// less then (TotalNumberOfPixels*Quantile) of the image area together are
// set to black, and all light pixels that occupy the same area together are
// set to white. Intermediate color values are linearly stretched.
// If Mode is emRGB, this process is done individually for each color channel,
// if Mode is emGray, gray values are used for calculation and all color
// channels are stretched equally.
// Note that modifications take place on BMP directly. For that purpose, BMP
// is converted to a PixelFormat of pf24bit.


// base type of a histogram
type PHistogram=^THistogram;
     THistogram=array[byte]of integer;

// extract histogram data from an existing bitmap
procedure GenerateRGBHistogram(BMP:TBitmap; var HR,HG,HB:THistogram);
procedure GenerateGrayHistogram(BMP:TBitmap; var Hist:THistogram);

// get the sum or the maximum value of a histogram
function GetHistogramSum(const Hist:THistogram):integer;
function GetHistogramMax(const Hist:THistogram):integer;

// render a histogram to a (256 pixels wide) bitmap (for analysis purposes)
procedure RenderRGBHistogram(const HR,HG,HB:THistogram; BMP:TBitmap);
procedure RenderGrayHistogram(const Hist:THistogram; BMP:TBitmap);

// get the intensity value that encompasses the given part of the histogram
function GetHistogramLowerQuantile(const Hist:THistogram; Sum:integer):integer;
function GetHistogramUpperQuantile(const Hist:THistogram; Sum:integer):integer;

// stretch the intensity levels of a bitmap so that Min->0 and Max->255
procedure StretchLevelsRGB(BMP:TBitmap; MinR,MaxR,MinG,MaxG,MinB,MaxB:integer);
procedure StretchLevelsGray(BMP:TBitmap; Min,Max:integer);

implementation
uses SysUtils, Classes;

procedure Equalize(BMP:TBitmap; Mode:TEqualizeMode; Quantile:real);
var HR,HG,HB:THistogram;
    QSum,RMin,RMax,GMin,GMax,BMin,BMax:integer;
begin
  if Mode=emRGB then begin
    GenerateRGBHistogram(BMP,HR,HG,HB);
    QSum:=round(GetHistogramSum(HG)*Quantile);
    RMin:=GetHistogramLowerQuantile(HR,QSum);
    RMax:=GetHistogramUpperQuantile(HR,QSum);
    GMin:=GetHistogramLowerQuantile(HG,QSum);
    GMax:=GetHistogramUpperQuantile(HG,QSum);
    BMin:=GetHistogramLowerQuantile(HB,QSum);
    BMax:=GetHistogramUpperQuantile(HB,QSum);
    StretchLevelsRGB(BMP,RMin,RMax,GMin,GMax,BMin,BMax);
  end else begin
    GenerateGrayHistogram(BMP,HG);
    QSum:=round(GetHistogramSum(HG)*Quantile);
    GMin:=GetHistogramLowerQuantile(HG,QSum);
    GMax:=GetHistogramUpperQuantile(HG,QSum);
    StretchLevelsGray(BMP,GMin,GMax);
  end;
end;

procedure GenerateRGBHistogram(BMP:TBitmap; var HR,HG,HB:THistogram);
var X,Y,MaxX:integer;
    L:PByteArray; P:^byte;
begin
  BMP.PixelFormat:=pf24bit;
  FillChar(HR,sizeof(THistogram),0);
  FillChar(HG,sizeof(THistogram),0);
  FillChar(HB,sizeof(THistogram),0);
  MaxX:=BMP.Width-1;
  for Y:=(BMP.Height-1) downto 0 do begin
    L:=BMP.ScanLine[Y]; P:=@L[0];
    for X:=MaxX downto 0 do begin
      inc(HB[P^]); inc(P);
      inc(HG[P^]); inc(P);
      inc(HR[P^]); inc(P);
    end;
  end;
end;

procedure GenerateGrayHistogram(BMP:TBitmap; var Hist:THistogram);
var X,Y,MaxX:integer;
    L:PByteArray; P:^byte;
begin
  BMP.PixelFormat:=pf24bit;
  FillChar(Hist,sizeof(THistogram),0);
  MaxX:=BMP.Width*3-1;
  for Y:=(BMP.Height-1) downto 0 do begin
    L:=BMP.ScanLine[Y]; P:=@L[0];
    for X:=MaxX downto 0 do begin
      inc(Hist[P^]); inc(P);
    end;
  end;
end;

function GetHistogramSum(const Hist:THistogram):integer;
var I:integer; P:^integer;
begin
  Result:=0;
  P:=@Hist[0];
  for I:=255 downto 0 do begin
    inc(Result,P^); inc(P);
  end;
end;

function GetHistogramMax(const Hist:THistogram):integer;
var I:integer; P:^integer;
begin
  Result:=0;
  P:=@Hist[0];
  for I:=255 downto 0 do begin
    if P^>Result then Result:=P^;
    inc(P);
  end;
end;

procedure RenderRGBHistogram(const HR,HG,HB:THistogram; BMP:TBitmap);
var X,ScaleDiv,ScaleMul:integer;
begin with BMP,Canvas do begin
  Brush.Style:=bsSolid;
  Brush.Color:=clBlack;
  FillRect(Rect(0,0,Width,Height));
  Pen.Style:=psSolid;
  Pen.Mode:=pmMerge;
  ScaleMul:=Height-1;
  ScaleDiv:=GetHistogramMax(HR);
  X:=GetHistogramMax(HG); if X>ScaleDiv then ScaleDiv:=X;
  X:=GetHistogramMax(HB); if X>ScaleDiv then ScaleDiv:=X;
  for X:=0 to 255 do begin
    Pen.Color:=clRed;
    MoveTo(X,ScaleMul-HR[X]*ScaleMul DIV ScaleDiv); LineTo(X,ScaleMul+1);
    Pen.Color:=clLime;
    MoveTo(X,ScaleMul-HG[X]*ScaleMul DIV ScaleDiv); LineTo(X,ScaleMul+1);
    Pen.Color:=clBlue;
    MoveTo(X,ScaleMul-HB[X]*ScaleMul DIV ScaleDiv); LineTo(X,ScaleMul+1);
  end;
end; end;

procedure RenderGrayHistogram(const Hist:THistogram; BMP:TBitmap);
var X,ScaleDiv,ScaleMul:integer;
begin with BMP,Canvas do begin
  Brush.Style:=bsSolid;
  Brush.Color:=clBlack;
  FillRect(Rect(0,0,Width,Height));
  Pen.Style:=psSolid;
  Pen.Color:=clWhite;
  Pen.Mode:=pmCopy;
  ScaleMul:=Height-1;
  ScaleDiv:=GetHistogramMax(Hist);
  for X:=0 to 255 do begin
    MoveTo(X,ScaleMul-Hist[X]*ScaleMul DIV ScaleDiv); LineTo(X,ScaleMul+1);
  end;
end; end;

function GetHistogramLowerQuantile(const Hist:THistogram; Sum:integer):integer;
var I,SumSoFar:integer;
begin
  SumSoFar:=0;
  for I:=0 to 255 do begin
    inc(SumSoFar,Hist[I]);
    if SumSoFar>Sum then begin
      Result:=I;
      exit;
    end;
  end;
  Result:=255;
end;

function GetHistogramUpperQuantile(const Hist:THistogram; Sum:integer):integer;
var I,SumSoFar:integer;
begin
  SumSoFar:=0;
  for I:=255 downto 0 do begin
    inc(SumSoFar,Hist[I]);
    if SumSoFar>Sum then begin
      Result:=I;
      exit;
    end;
  end;
  Result:=0;
end;

procedure StretchLevelsRGB(BMP:TBitmap; MinR,MaxR,MinG,MaxG,MinB,MaxB:integer);
var X,Y,MaxX,I:integer;
    L:PByteArray; P:^byte;
    DiffR,DiffG,DiffB:integer;
begin
  BMP.PixelFormat:=pf24bit;
  DiffR:=MaxR-MinR; if DiffR=0 then DiffR:=1;
  DiffG:=MaxG-MinG; if DiffG=0 then DiffG:=1;
  DiffB:=MaxB-MinB; if DiffB=0 then DiffB:=1;
  MaxX:=BMP.Width-1;
  for Y:=(BMP.Height-1) downto 0 do begin
    L:=BMP.ScanLine[Y]; P:=@L[0];
    for X:=MaxX downto 0 do begin
      I:=MinB+((P^-MinB)*255 DIV DiffB);
      if I<0 then P^:=0 else if I>255 then P^:=255 else P^:=I; inc(P);
      I:=MinG+((P^-MinG)*255 DIV DiffG);
      if I<0 then P^:=0 else if I>255 then P^:=255 else P^:=I; inc(P);
      I:=MinR+((P^-MinR)*255 DIV DiffR);
      if I<0 then P^:=0 else if I>255 then P^:=255 else P^:=I; inc(P);
    end;
  end;
end;

procedure StretchLevelsGray(BMP:TBitmap; Min,Max:integer);
var X,Y,MaxX,I:integer;
    L:PByteArray; P:^byte;
    Diff:integer;
begin
  BMP.PixelFormat:=pf24bit;
  Diff:=Max-Min; if Diff=0 then Diff:=1;
  MaxX:=BMP.Width*3-1;
  for Y:=(BMP.Height-1) downto 0 do begin
    L:=BMP.ScanLine[Y]; P:=@L[0];
    for X:=MaxX downto 0 do begin
      I:=Min+((P^-Min)*255 DIV Diff);
      if I<0 then P^:=0 else if I>255 then P^:=255 else P^:=I; inc(P);
    end;
  end;
end;

end.
