unit ScaleBMP;
// BMP Scaler Toolkit (C) 2002 Martin J. Fiedler <martin.fiedler@gmx.net>
// This is free software. You may freely redistribute and modify it.
interface
uses SysUtils,Graphics;

procedure Scale(Source,Dest:TBitmap);

// The usage is simple: create two bitmaps of any size you want, call Scale()
// and you'll have a scaled copy of Source in Dest. Please note that both
// bitmaps' PixelFormat is set to pf24bit.

// Scaling of the X and Y axis is independent, i.e. it is possible to reduce
// the image width when increasing the height or vice-versa.

// The downscaling algorithm is a simple weighed linear averaging algorithm,
// which turns out to be exactly the same as Paint Shop Pro uses (except some
// roundoff errors).

// The upscaling algorithm is a simple (bi-)linear interpolation, just like it's
// done by modern graphics accelerator chips.

// The speed of the algorithms is far from optimal: Although the TBitmap-related
// parts are to some extent optimized (ScanLine used, already calculated lines
// are cached), most computations are done in single-precision floating-point
// arithmetic which results in typically half the speed compared to Paint Shop
// Pro, for example. (Maybe it gets better if Borland writes a Delphi compiler
// supporting SSE some day :)

implementation

procedure Scale(Source,Dest:TBitmap);

type TLine=array of single;
type TCacheLine=record
                  Line:TLine;
                  CachedLine:integer;
                end;
const EPSILON=0.001;

var SizeX,SizeY,NewX,NewY,LineSize:integer;
var dX,dY:single;
var XScaler:(Copy,Up,Down);
var Sum:TLine; LineCache:array[1..2]of TCacheLine;

  procedure InitXCopy;
  begin
    XScaler:=Copy;
  end;
  procedure DoXCopy(var Line:TLine;Y:integer);
  var L:PByteArray; PS:^byte; PD:^single; X:integer;
  begin
    L:=Source.ScanLine[Y];
    PS:=@L[0]; PD:=@Line[0];
    for X:=LineSize downto 1 do begin
      PD^:=PS^; inc(PS); inc(PD);
    end;
  end;

  procedure InitXScaleDown;
  begin
    dX:=SizeX/NewX;
    XScaler:=Down;
  end;
  procedure DoXScaleDown(var Line:TLine;Y:integer);
  var NX,SrcX:integer;
      StartX,EndX:single;
      NextX,DiffX:single;
      SR,SG,SB,Count:single;
      PixelFinished:boolean;
      L:PByteArray; PD:^single;
  begin
    L:=Source.ScanLine[Y]; PD:=@Line[0];
    StartX:=0.0;
    for NX:=0 to (NewX-1) do begin
      EndX:=dX*(NX+1);
      SR:=0.0; SG:=0.0; SB:=0.0; Count:=0.0;
      PixelFinished:=false;
      repeat
        SrcX:=trunc(StartX);
        NextX:=SrcX+1.0;
        if NextX>=EndX then begin
          NextX:=EndX;
          PixelFinished:=true;
        end;
        DiffX:=NextX-StartX;
        if (DiffX>EPSILON) AND (DiffX<SizeX) then begin
          SB:=SB+DiffX*L[3*SrcX];
          SG:=SG+DiffX*L[3*SrcX+1];
          SR:=SR+DiffX*L[3*SrcX+2];
          Count:=Count+DiffX;
        end;
        StartX:=NextX;
      until PixelFinished;
      PD^:=SB/Count; inc(PD);
      PD^:=SG/Count; inc(PD);
      PD^:=SR/Count; inc(PD);
      StartX:=EndX;
    end;
  end;

  procedure InitXScaleUp;
  begin
    dX:=(SizeX-1)/(NewX-1);
    XScaler:=Up;
  end;
  procedure DoXScaleUp(var Line:TLine;Y:integer);
  var NX,IX:integer;
      PX,SubX,IBias:single;
      L:PByteArray; PD:^single;
  begin
    L:=Source.ScanLine[Y]; PD:=@Line[0];
    for NX:=0 to (NewX-1) do begin
      PX:=NX*dX;
      IX:=trunc(PX);
      if (IX=PX) OR (IX+1>=SizeX) then begin
        PD^:=L[3*IX]; inc(PD);
        PD^:=L[3*IX+1]; inc(PD);
        PD^:=L[3*IX+2]; inc(PD);
      end else begin
        SubX:=PX-IX;
        IBias:=1.0-SubX;
        PD^:=IBias*L[3*IX]  +SubX*L[3*IX+3]; inc(PD);
        PD^:=IBias*L[3*IX+1]+SubX*L[3*IX+4]; inc(PD);
        PD^:=IBias*L[3*IX+2]+SubX*L[3*IX+5]; inc(PD);
      end;
    end;
  end;

  function GetLine(Y:integer):integer;
  var MinD,MinDAt,I,D:integer;
  begin
    MinD:=maxint; MinDAt:=1;
    for I:=1 to 2 do begin
      D:=abs(Y-LineCache[I].CachedLine);
      if D<MinD then begin
        MinD:=D; MinDAt:=I;
      end;
    end;
    if MinD=0 then begin
      Result:=MinDAt;
      exit;
    end;
    Result:=3-MinDAt;
    LineCache[Result].CachedLine:=Y;
    case XScaler of
      Copy:DoXCopy(LineCache[Result].Line,Y);
      Down:DoXScaleDown(LineCache[Result].Line,Y);
      Up:DoXScaleUp(LineCache[Result].Line,Y);
    end;
  end;

  procedure ClearLine(var Sum:TLine);
  var X:integer; PD:^single;
  begin
    PD:=@Sum[0];
    for X:=LineSize downto 1 do begin
      PD^:=0.0; inc(PD);
    end;
  end;

  procedure CopyLine(var Line,Sum:TLine;Bias:single);
  var X:integer; PS,PD:^single;
  begin
    PS:=@Line[0]; PD:=@Sum[0];
    for X:=LineSize downto 1 do begin
      PD^:=PS^*Bias; inc(PS); inc(PD);
    end;
  end;

  procedure AddLine(var Line,Sum:TLine;Bias:single);
  var X:integer; PS,PD:^single;
  begin
    PS:=@Line[0]; PD:=@Sum[0];
    for X:=LineSize downto 1 do begin
      PD^:=PD^+PS^*Bias; inc(PS); inc(PD);
    end;
  end;

  procedure PutLine(var Sum:TLine;Bias:single;Y:integer);
  var L:PByteArray; PS:^single; PD:^byte; X,I:integer;
  begin
    L:=Dest.ScanLine[Y];
    PS:=@Sum[0]; PD:=@L[0];
    if Bias=1.0 then
      for X:=LineSize downto 1 do begin
        I:=round(PS^);
        if I<0 then I:=0; if I>255 then I:=255;
        PD^:=I;
        inc(PS); inc(PD);
      end
    else
      for X:=LineSize downto 1 do begin
        I:=round(PS^/Bias);
        if I<0 then I:=0; if I>255 then I:=255;
        PD^:=I;
        inc(PS); inc(PD);
      end;
  end;

  procedure DoYCopy;
  var Y,LI:integer;
  begin
    for Y:=0 to (NewY-1) do begin
      LI:=GetLine(Y);
      PutLine(LineCache[LI].Line,1.0,Y);
    end;
  end;

  procedure DoYScaleDown;
  var NY,SrcY:integer;
      StartY,EndY:single;
      NextY,DiffY:single;
      Count:single;
      LineFinished:boolean;
      LI:integer;
  begin
    dY:=SizeY/NewY;
    StartY:=0.0;
    for NY:=0 to (NewY-1) do begin
      EndY:=dY*(NY+1);
      ClearLine(Sum); Count:=0.0;
      LineFinished:=false;
      repeat
        SrcY:=trunc(StartY);
        NextY:=SrcY+1.0;
        if NextY>=EndY then begin
          NextY:=EndY;
          LineFinished:=true;
        end;
        DiffY:=NextY-StartY;
        if (DiffY>EPSILON) AND (SrcY<SizeY) then begin
          LI:=GetLine(SrcY);
          AddLine(LineCache[LI].Line,Sum,DiffY);
          Count:=Count+DiffY;
        end;
        StartY:=NextY;
      until LineFinished;
      PutLine(Sum,Count,NY);
      StartY:=EndY;
    end;
  end;

  procedure DoYScaleUp;
  var NY,IY:integer;
      PY,SubY:single;
      LI:integer;
  begin
    dY:=(SizeY-1)/(NewY-1);
    for NY:=0 to (NewY-1) do begin
      PY:=dY*NY;
      IY:=trunc(PY);
      if (IY=PY) OR (IY+1>=SizeY) then begin
        LI:=GetLine(IY);
        PutLine(LineCache[LI].Line,1.0,NY);
      end else begin
        SubY:=PY-IY;
        LI:=GetLine(IY);
        CopyLine(LineCache[LI].Line,Sum,1.0-SubY);
        LI:=GetLine(IY+1);
        AddLine(LineCache[LI].Line,Sum,SubY);
        PutLine(Sum,1.0,NY);
      end;
    end;
  end;

begin
  if not(Assigned(Source)) OR not(Assigned(Dest)) then exit;
  with Source do begin
    PixelFormat:=pf24bit;
    SizeX:=Width; SizeY:=Height;
  end;
  with Dest do begin
    PixelFormat:=pf24bit;
    NewX:=Width; NewY:=Height;
  end;
  if (NewX=SizeX) AND (NewY=SizeY) then begin
    Dest.Canvas.Draw(0,0,Source);
    exit;
  end;
       if NewX=SizeX then InitXCopy
  else if NewX<SizeX then InitXScaleDown
  else                    InitXScaleUp;
  LineSize:=NewX*3;
  SetLength(LineCache[1].Line,LineSize); LineCache[1].CachedLine:=-1;
  SetLength(LineCache[2].Line,LineSize); LineCache[2].CachedLine:=-1;
  SetLength(Sum,LineSize);
       if NewY=SizeY then DoYCopy
  else if NewY<SizeY then DoYScaleDown
  else                    DoYScaleUp;
  Finalize(Sum);
  Finalize(LineCache[2].Line);
  Finalize(LineCache[1].Line);
end;

end.
